@@ -93,7 +93,8 @@ def improve(comment, constrainer, scorer, start_x, ftol=1e-4, initial_accuracy=0
         i = 0
         jobs = [ ]
         
-        pool_size = int(len(best)**1.5+1)
+        pool_size = int(len(best)*5)
+        print len(best),'parameters, pool size', pool_size
 
         currents = [ (best, best_score) ]
         
