@@ -3,19 +3,68 @@ import os
 
 import profile, shape
 
+def plan_segments(cuts, length_ratio):
+    plans = [ ]
+    for i in xrange(1<<len(cuts)):
+        values = [ (i>>j)&1 for j in xrange(len(cuts)) ]
+        upper = [0.0] + [ cuts[j] for j,k in enumerate(values) if k == 1] + [ 1.0 ]
+        #lower = [0.0] + [ cuts[j] for j,k in enumerate(values) if k == 2] + [ 1.0 ]
+        
+        lower = [0.0] + [ (upper[j]+upper[j+1])*0.5 for j in xrange(len(upper)-1) ] + [1.0]
+        j = 0
+        while j < len(lower)-2:
+            if lower[j+2]-lower[j] <= length_ratio:
+                del lower[j+1]
+            else:
+                j += 1
+        
+        good = True
+        for seg in (upper,lower):
+            for j in xrange(len(seg)-1):
+                if seg[j+1]-seg[j] > length_ratio:
+                    good = False
+                    break
+            if not good: break
+        if not good: continue
+        plans.append(( upper, lower ))
+    
+    assert plans, 'Couldn\'t find a way to segment for milling.'    
+    
+    def scorer((upper,lower)):
+        worst = 1e30
+        for a in upper[1:-1]:
+            for b in lower[1:-1]:
+                worst = min(worst, abs(a-b))
+        return len(upper)+len(lower),len(upper),-worst
+    
+    best = min(plans, key=scorer)
+    print 'Segmentation:'
+    print ' '.join('%3.2f' % i for i in best[0])
+    print ' '.join('%3.2f' % i for i in best[1])
+    print
+    
+    return best
+                
+
+
 class Packable(object):
     def __init__(self, shapes, rotation, dilation, use_upper=True):
-        self.shapes = [ item.copy() for item in shapes ]
-        for item in self.shapes:
-            item.rotate(0,0,1, rotation)
+        if not rotation:
+            self.shapes = shapes
+        else:
+            self.shapes = [ item.copy() for item in shapes ]
+            for item in self.shapes:
+                item.rotate(0,0,1, rotation)
         
         extent = self.shapes[0].extent()
         for item in self.shapes:
             item.move(-extent.xmin,-extent.ymin,0)
         self.extent = self.shapes[0].extent()
         
-        self.mask = self.shapes[0].polygon_mask()
-        self.mask = shape.create_polygon_2(self.mask.loop(holes=False))
+        mask = self.shapes[0].polygon_mask()
+        loop = mask.loop(holes=False)
+        self.mask = shape.create_polygon_2(loop)
+        self.centroid = loop.centroid
 
         self.dilated_mask = self.mask.offset_curve(dilation, 16)
         self.dilated_mask = shape.create_polygon_2(self.dilated_mask.loop(holes=False))
@@ -57,30 +106,18 @@ class Pack(object):
 
     def render(self, bit_diameter, bit_pad):
         xsize = self.xsize
-        ysize = self.ysize #max( y+packable.extent.ymax for x,y,packable in self.items )+bit_diameter*2
+        ysize = self.ysize
         
         a = bit_pad
         b = bit_pad + self.zsize/10.0
         pad_cone = shape.extrude_profile(profile.Profile([0,self.zsize],[a*2,b*2],[a*2,b*2]), cross_section=lambda d: shape.circle(d,16))
+        pad_cylinder = shape.extrude_profile(profile.Profile([0,self.zsize+bit_diameter],[a*2,a*2],[a*2,a*2]), cross_section=lambda d: shape.circle(d,16))
         
         pad = 0.5 + self.zsize/10.0
         
-#        dilations = [
-##            (z, shape.circle(2.0*( 2.0+z/20.0 )).mask(self.res))
-##            for z in range(1)
-#            (z, mask.ones(0,0,1,1) if not z else 
-#                shape.circle(2.0*( z/10.0 )).mask(self.res))
-#            for z in range(0,int(self.zsize),5)
-#        ]
-
-        #dilator = shape.circle(2* bit_diameter).mask(res)
-        #hole_mask = mask.zero()
-        #for x,y,packable in self.items:
-        #    if packable.use_upper:
-        #        hole_mask = hole_mask | packable.dilated_mask.offset(int(x*self.res+0.5),int(y*self.res+0.5))
-        #hole_mask = hole_mask.close(dilator)
-        
         upper = shape.block(-pad,xsize+pad,-pad,ysize+pad,0,self.zsize)
+        
+        print 'Cut holes'
 
         for i,(x,y,packable) in enumerate(self.items):
             if packable.use_upper:
@@ -89,13 +126,15 @@ class Pack(object):
                 minsum = flat.minkowski_sum(pad_cone)
                 upper.remove( minsum )
 
+        print 'Put things in them'
+
         for x,y,packable in self.items:
             if packable.use_upper:
                 temp = packable.shapes[0].copy()
                 temp.move(x,y,0)
                 upper.add(temp)
             
-        bol = int(self.xsize / 30.0)
+        bol = int(self.xsize / 20.0)
         bol_width = 1.0
             # 0.5 seemed to allow shifting on a sop flute
         bol_height = 1.0 
@@ -103,11 +142,12 @@ class Pack(object):
             # contour finishing (x-scan of horizontal-like surfaces)
         xmins = [ x for x,y,packable in self.items ]
         xmaxs = [ x+packable.extent.xmax for x,y,packable in self.items ]
-        margin = bit_diameter*2
+        margin = bol_width+bit_diameter+0.25
         for i in range(bol):
             pos = (i+0.5)*xsize/bol
             
-            if any( pos-margin <= item <= pos+margin for item in xmins+xmaxs ):
+            if any( pos-margin <= item <= pos+bol_width for item in xmins ) or \
+               any( pos-bol_width <= item <= pos+margin for item in xmins ):
                 continue
             
             upper.add(shape.block(
@@ -117,30 +157,54 @@ class Pack(object):
                 0,bol_height,
             ))     
         
+        
+        print 'Underside'
+        
         # Cut the bottom of upper with lower, to depth    bit_diameter
         lower = shape.block(-pad,xsize+pad,-pad,ysize+pad,bit_diameter,self.zsize)
         lower.add(upper)
+        
+        print 'Remove bores'
 
         for x,y,packable in self.items:
             temp = packable.shapes[1].copy()
             temp.move(x,y,0)
+
+            flat = packable.mask.to_3()
+            flat.move(x,y,0)
+            minsum = flat.minkowski_sum(pad_cylinder)
+            temp.clip(minsum)
+
             lower.remove(temp)
         
         lower.rotate(0,1,0, 180)
         
         return lower, upper
+    
+    def render_print(self):
+        result = None
+        for x,y,packable in self.items:
+            temp = packable.shapes[0].copy()
+            temp.move(x,y,0)
+            if result is None:
+                result = temp
+            else:
+                result.add(temp)
+        return result
 
 
-def make_segment(instrument, top, low, high, radius, pad=0.0, clip_half=True):
-    if not clip_half:
-       y1,y2 = -radius,radius
-    elif top:
-       y1,y2 = 0,radius
-    else:
-       y1,y2 = -radius,0
-    clip = shape.block(-radius,radius, y1,y2, low-pad,high+pad)
+def make_segment(instrument, top, low, high, radius, clip=True):
+    #if not clip_half:
+    #   y1,y2 = -radius,radius
     segment = instrument.copy()
-    segment.clip(clip)
+    
+    if clip:
+        if top:
+           y1,y2 = 0,radius
+        else:
+           y1,y2 = -radius,0
+        clipper = shape.block(-radius,radius, y1,y2, low,high)
+        segment.clip(clipper)
     #Closed flute is meant to be like this
     #assert abs(segment.size()[2] - (high-low+pad*2)) < 1e-3, 'Need more padding for construction'
     segment.move(0,0,-low)
@@ -153,47 +217,74 @@ def make_segment(instrument, top, low, high, radius, pad=0.0, clip_half=True):
     return segment
 
 
-def make_segments(instrument, length, radius, top_fractions, bottom_fractions, pad=0.0, clip_half=True):
+def make_segments(instrument, length, radius, top_fractions, bottom_fractions, clip=True):
     parts = [ ]
-    lengths = [ ]
     z = [ item*length for item in top_fractions ]
     for i in range(len(top_fractions)-1):
-        lengths.append(z[i+1]-z[i])    
         parts.append(make_segment(instrument, True, 
                      z[i],z[i+1], 
-                     radius, pad, clip_half))
+                     radius, clip))
 
     z = [ item*length for item in bottom_fractions ]
     for i in range(len(bottom_fractions)-1):    
-        lengths.append(z[i+1]-z[i])    
         parts.append(make_segment(instrument, False, 
                      z[i],z[i+1], 
-                     radius, pad, clip_half))
+                     radius, clip))
 
-    return parts, lengths
+    return parts
 
 
 def deconstruct(outer, bore, top_fractions, bottom_fractions, 
-                bit_diameter, dilation):
+                bit_diameter, dilation, block_zsize):
     """ Object must be large enough to include end padding """
     xsize,ysize,zsize = outer.size()
     radius = max(xsize,ysize)*2
     
     length = zsize
     
-    outers, lengths = make_segments(outer, length, radius, top_fractions, bottom_fractions, 0)
-    bores, _ = make_segments(bore, length, radius, top_fractions, bottom_fractions, bit_diameter, False)
+    outers = make_segments(outer, length, radius, top_fractions, bottom_fractions)
+    bores  = make_segments(bore, length, radius, top_fractions, bottom_fractions, False)
 
-    return [ [ Packable([outer, bore], 0, dilation),
-                Packable([outer, bore], 180, dilation) ]
-              for outer, bore in zip(outers, bores) ]
+    result = [ ]
+    for outer, bore in zip(outers, bores):
+        extent = outer.extent()
+        
+        b = bit_diameter
+        clipper = shape.block(extent.xmin-b,extent.xmax+b,
+                              extent.ymin-b,extent.ymax+b,
+                              extent.zmin,extent.zmax+b)
+        bore.clip(clipper)
+        
+        n = 1
+        while extent.zmax > block_zsize*n:
+            n += 1
+        #thickness = extent.zmax/n
+        thickness = block_zsize
+        
+        block = shape.block(extent.xmin-1,extent.xmax+1,extent.ymin-1,extent.ymax+1,0,thickness)
+        for i in xrange(n):
+            temp_outer = outer.copy()
+            temp_bore = bore.copy()
+            temp_outer.move(0,0,-i*thickness)
+            temp_bore.move(0,0,-i*thickness)
+            temp_outer.clip(block)
+            #temp_bore.clip(block)
+            result.append([
+                Packable([temp_outer,temp_bore], 0, dilation),
+                Packable([temp_outer,temp_bore], 180, dilation),
+                ])
+    return result
+
+    #return [ [ Packable([outer, bore], 0, dilation),
+    #            Packable([outer, bore], 180, dilation) ]
+    #          for outer, bore in zip(outers, bores) ]
 
 
-def pack(packables, xsize, ysize, zsize, dilation):
+def mill_template(xsize, ysize, zsize, dilation):
     template = Pack(xsize, ysize, zsize)
     
-    peg_diameter = 6
-    block_radius = 3.5
+    peg_diameter = 6.0
+    block_radius = 4.0
     hole = shape.prism(min(zsize,peg_diameter*2.0), peg_diameter)   # *3 is too deep
     hole_block = shape.block(-block_radius,block_radius,-block_radius,block_radius,0,zsize)
     hole_packable = Packable([hole_block,hole], 0, 0.0, use_upper=False)
@@ -205,6 +296,26 @@ def pack(packables, xsize, ysize, zsize, dilation):
     template.put(0,ysize-block_radius*3,hole_packable)
     template.put(xsize-block_radius*2,ysize-block_radius*3,hole_packable)
     
+    return template
+
+def mill_squish(pack):
+    x1, y, hole1 = pack.items.pop(2)
+    x2, y, hole2 = pack.items.pop(2)
+    del pack.masks[2:4]
+    y_new = y
+    step = 1.0
+        
+    y_min = max( item[1]+item[2].extent.ymax for item in pack.items )
+    
+    while y_new-step >= y_min and pack.valid(x1,y_new-step,hole1) and pack.valid(x2,y_new-step,hole2):
+        y_new -= step
+        
+    pack.put(x1,y_new,hole1)
+    pack.put(x2,y_new,hole2)
+    pack.ysize -= y-y_new
+
+
+def pack(template, packables, aspect_goal=1.0):    
     packs = [ template.copy() ]
     pack = packs[-1]
     
@@ -212,8 +323,8 @@ def pack(packables, xsize, ysize, zsize, dilation):
     step_back = 0.1
 
     points = [ ]
-    for y in xrange(int(ysize/step)):
-        for x in xrange(int(xsize/step)):
+    for y in xrange(int(template.ysize/step)):
+        for x in xrange(int(template.xsize/step)):
             points.append((x*step,y*step))
     points.sort(key=lambda item: (-item[1],-item[0]))
     
@@ -221,28 +332,33 @@ def pack(packables, xsize, ysize, zsize, dilation):
     while todo:
         try:
             queue = [
-                (i,x,y)
+                (i,j,x,y)
                 for x,y in points
-                for i in range(len(todo))
+                for i in xrange(len(todo))
+                for j in xrange(len(todo[i]))
             ]
         
             #for i in range(len(todo)):
             #    for x,y in points:
             
             def sorter(item):
-                i,x,y = item
-                return (
-                    y >= block_radius*2,
-                    -todo[i][0].extent.xmax,
-                    y,
-                    x,
-                )
+                i,j,x,y = item
+                xm = todo[i][j].extent.xmax
+                ym = todo[i][j].extent.ymax
+                xc, yc = todo[i][j].centroid
+                return (((aspect_goal*(x+xc))**2+(y+yc)**2) - ((aspect_goal*xm)**2+(ym)**2), y, x)
+                #return (max(aspect_goal*(x+xc),(y+yc)) -max(aspect_goal*xc,yc), y, x)
+                #return (
+                #    y > 0,
+                #    -todo[i][0].extent.xmax,
+                #    y,
+                #    x,
+                #)
             queue.sort(key=sorter)
             
-            for i,x,y in queue:
-                for j in range(len(todo[i])):
-                    if pack.valid(x,y,todo[i][j]):
-                        raise StopIteration()
+            for i,j,x,y in queue:
+                if pack.valid(x,y,todo[i][j]):
+                    raise StopIteration()
 
             assert len(pack.items) > len(template.items), 'This is never going to work'
             pack = template.copy()
@@ -276,19 +392,24 @@ def save_packs(packs, save, bit_diameter, pad):
     return shapes
 
 
-def cut_and_pack(outer, bore, top_fractions, bottom_fractions, xsize, ysize, zsize, bit_diameter, save):
-    pad = bit_diameter*1.25
+def cut_and_pack(outer, bore, top_fractions, bottom_fractions, xsize, ysize, zsize, bit_diameter, save, extra_packables=[]):
+    pad = bit_diameter*1.2
     dilation = pad
     
-    packables = deconstruct(outer, bore, top_fractions, bottom_fractions, bit_diameter, dilation)
-
-    #for i, item in enumerate([ a for b in packables for a in b ]):
-    #    mask.write_pbm(open('/tmp/mask%d.pbm'%i,'w'), item.dilated_mask.data)
+    packables = deconstruct(outer, bore, top_fractions, bottom_fractions, bit_diameter, dilation, zsize)
+    packables.extend(extra_packables)
 
     for item in packables:
-        print('%.1fmm x %.1fmm x %.1fmm' % (item[0].extent.xmax,item[0].extent.ymax,item[0].extent.zmax))
+        print '%.1fmm x %.1fmm x %.1fmm' % (item[0].extent.xmax,item[0].extent.ymax,item[0].extent.zmax)
 
-    packs = pack(packables, xsize, ysize, zsize, dilation)
+    print 'Packing'
+
+    template = mill_template(xsize + pad*2, ysize + pad*2, zsize, dilation)
+    packs = pack(template, packables, float(ysize)/xsize)
+    for item in packs:
+        mill_squish(item)
+    
+    print 'Rendering'
     
     return save_packs(packs, save, bit_diameter, pad)
 
